@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2001  Internet Software Consortium.
+ * Copyright (C) 1999-2002  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -15,10 +15,11 @@
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: update.c,v 1.79.2.3 2001/02/22 19:33:12 gson Exp $ */
+/* $Id: update.c,v 1.88.2.3 2002/02/08 03:57:15 marka Exp $ */
 
 #include <config.h>
 
+#include <isc/print.h>
 #include <isc/string.h>
 #include <isc/taskpool.h>
 #include <isc/util.h>
@@ -32,8 +33,10 @@
 #include <dns/journal.h>
 #include <dns/message.h>
 #include <dns/nxt.h>
+#include <dns/rdataclass.h>
 #include <dns/rdataset.h>
 #include <dns/rdatasetiter.h>
+#include <dns/soa.h>
 #include <dns/ssu.h>
 #include <dns/view.h>
 #include <dns/zone.h>
@@ -55,28 +58,14 @@
 /**************************************************************************/
 
 /*
- * Convenience macro of common isc_log_write() arguments
- * to use in reportings server errors.
+ * Log level for tracing dynamic update protocol requests.
  */
-#define UPDATE_ERROR_LOGARGS \
-	ns_g_lctx, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE, \
-	ISC_LOG_ERROR
+#define LOGLEVEL_PROTOCOL	ISC_LOG_INFO
 
 /*
- * Convenience macro of common isc_log_write() arguments
- * to use in tracing dynamic update protocol requests.
+ * Log level for low-level debug tracing.
  */
-#define UPDATE_PROTOCOL_LOGARGS \
-	ns_g_lctx, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE, \
-	ISC_LOG_INFO
-
-/*
- * Convenience macro of common isc_log_write() arguments
- * to use in low-level debug tracing.
- */
-#define UPDATE_DEBUG_LOGARGS \
-	ns_g_lctx, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE, \
-	ISC_LOG_DEBUG(8)
+#define LOGLEVEL_DEBUG 		ISC_LOG_DEBUG(8)
 
 /*
  * Check an operation for failure.  These macros all assume that
@@ -111,8 +100,8 @@
 #define FAILC(code, msg) \
 	do {							\
 		result = (code);				\
-		isc_log_write(UPDATE_PROTOCOL_LOGARGS,		\
-			      "dynamic update failed: %s (%s)",	\
+		update_log(client, zone, LOGLEVEL_PROTOCOL,   	\
+			      "update failed: %s (%s)",		\
 		      	      msg, isc_result_totext(result));	\
 		if (result != ISC_R_SUCCESS) goto failure;	\
 	} while (0)
@@ -125,8 +114,8 @@
 #define FAILS(code, msg) \
 	do {							\
 		result = (code);				\
-		isc_log_write(UPDATE_PROTOCOL_LOGARGS,		\
-			      "dynamic update error: %s: %s", 	\
+		update_log(client, zone, LOGLEVEL_PROTOCOL,		\
+			      "error: %s: %s", 			\
 			      msg, isc_result_totext(result));	\
 		if (result != ISC_R_SUCCESS) goto failure;	\
 	} while (0)
@@ -161,6 +150,40 @@ static isc_result_t send_forward_event(ns_client_t *client, dns_zone_t *zone);
 static void forward_done(isc_task_t *task, isc_event_t *event);
 
 /**************************************************************************/
+
+static void
+update_log(ns_client_t *client, dns_zone_t *zone,
+	   int level, const char *fmt, ...) ISC_FORMAT_PRINTF(4, 5);
+
+static void
+update_log(ns_client_t *client, dns_zone_t *zone,
+	   int level, const char *fmt, ...)
+{
+	va_list ap;
+	char message[4096];
+	char namebuf[DNS_NAME_FORMATSIZE];
+	char classbuf[DNS_RDATACLASS_FORMATSIZE];
+
+	if (client == NULL || zone == NULL)
+		return;
+
+	if (isc_log_wouldlog(ns_g_lctx, level) == ISC_FALSE)
+		return;
+
+	dns_name_format(dns_zone_getorigin(zone), namebuf,
+			sizeof(namebuf));
+	dns_rdataclass_format(dns_zone_getclass(zone), classbuf,
+			      sizeof(classbuf));
+
+	va_start(ap, fmt);
+	vsnprintf(message, sizeof message, fmt, ap);
+	va_end(ap);
+
+	ns_client_log(client, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE,
+		      level, "updating zone '%s/%s': %s",
+		      namebuf, classbuf, message);
+}
+
 /*
  * Update a single RR in version 'ver' of 'db' and log the
  * update in 'diff'.
@@ -681,10 +704,10 @@ temp_check_rrset(dns_difftuple_t *a, dns_difftuple_t *b) {
  */
 static int
 temp_order(const void *av, const void *bv) {
-	dns_difftuple_t * const *ap = av;
-	dns_difftuple_t * const *bp = bv;
-	dns_difftuple_t *a = *ap;
-	dns_difftuple_t *b = *bp;
+	dns_difftuple_t const * const *ap = av;
+	dns_difftuple_t const * const *bp = bv;
+	dns_difftuple_t const *a = *ap;
+	dns_difftuple_t const *b = *bp;
 	int r;
 	r = dns_name_compare(&a->name, &b->name);
 	if (r != 0)
@@ -1268,10 +1291,10 @@ non_nxt_rrset_exists(dns_db_t *db, dns_dbversion_t *ver,
  */
 static int
 name_order(const void *av, const void *bv) {
-	dns_difftuple_t * const *ap = av;
-	dns_difftuple_t * const *bp = bv;
-	dns_difftuple_t *a = *ap;
-	dns_difftuple_t *b = *bp;
+	dns_difftuple_t const * const *ap = av;
+	dns_difftuple_t const * const *bp = bv;
+	dns_difftuple_t const *a = *ap;
+	dns_difftuple_t const *b = *bp;
 	return (dns_name_compare(&a->name, &b->name));
 }
 
@@ -1394,6 +1417,7 @@ next_active(dns_db_t *db, dns_dbversion_t *ver, dns_name_t *oldname,
 
 /*
  * Add a NXT record for "name", recording the change in "diff".
+ * The existing NXT is removed.
  */
 static isc_result_t
 add_nxt(dns_db_t *db, dns_dbversion_t *ver, dns_name_t *name, dns_diff_t *diff)
@@ -1409,7 +1433,6 @@ add_nxt(dns_db_t *db, dns_dbversion_t *ver, dns_name_t *name, dns_diff_t *diff)
 	dns_fixedname_init(&fixedname);
 	target = dns_fixedname_name(&fixedname);
 
-
 	/*
 	 * Find the successor name, aka NXT target.
 	 */
@@ -1424,7 +1447,12 @@ add_nxt(dns_db_t *db, dns_dbversion_t *ver, dns_name_t *name, dns_diff_t *diff)
 	dns_db_detachnode(db, &node);
 
 	/*
-	 * Create a diff tuple, update the database, and record the change.
+	 * Delete the old NXT and record the change.
+	 */
+	CHECK(delete_if(true_p, db, ver, name, dns_rdatatype_nxt, 0,
+			NULL, diff));
+	/*
+	 * Add the new NXT and record the change.
 	 */
 	CHECK(dns_difftuple_create(diff->mctx, DNS_DIFFOP_ADD, name,
 				   3600,	/* XXXRTH */
@@ -1775,9 +1803,6 @@ update_signatures(isc_mem_t *mctx, dns_db_t *db, dns_dbversion_t *oldver,
 			 * there is other data, and if there is other data,
 			 * there are other SIGs.
 			 */
-			CHECK(delete_if(true_p, db, newver, &t->name,
-					dns_rdatatype_nxt, 0,
-					NULL, &nxt_diff));
 			CHECK(add_nxt(db, newver, &t->name, &nxt_diff));
 		}
 	}
@@ -2093,19 +2118,28 @@ update_action(isc_task_t *task, isc_event_t *event) {
 		FAILC(result, "'RRset exists (value dependent)' "
 		      "prerequisite not satisfied");
 
-	isc_log_write(UPDATE_DEBUG_LOGARGS, "prerequisites are OK");
+	update_log(client, zone, LOGLEVEL_DEBUG,
+		   "prerequisites are OK");
 
 	/*
 	 * Check Requestor's Permissions.  It seems a bit silly to do this
 	 * only after prerequisite testing, but that is what RFC2136 says.
 	 */
-	if (ssutable == NULL)
-		CHECK(ns_client_checkacl(client, "update",
+	if (ssutable == NULL) {
+		char msg[DNS_RDATACLASS_FORMATSIZE + DNS_NAME_FORMATSIZE
+			 + sizeof("update '/'")];
+		ns_client_aclmsg("update", zonename, client->view->rdclass,
+                                 msg, sizeof(msg));
+		CHECK(ns_client_checkacl(client, msg,
 					 dns_zone_getupdateacl(zone),
 					 ISC_FALSE, ISC_LOG_ERROR));
-	else if (client->signer == NULL) {
+	} else if (client->signer == NULL) {
 		/* This gets us a free log message. */
-		CHECK(ns_client_checkacl(client, "update", NULL, ISC_FALSE,
+		char msg[DNS_RDATACLASS_FORMATSIZE + DNS_NAME_FORMATSIZE
+			 + sizeof("update '/'")];
+		ns_client_aclmsg("update", zonename, client->view->rdclass,
+                                 msg, sizeof(msg));
+		CHECK(ns_client_checkacl(client, msg, NULL, ISC_FALSE,
 					 ISC_LOG_ERROR));
 	}
 
@@ -2149,10 +2183,9 @@ update_action(isc_task_t *task, isc_event_t *event) {
 				FAILC(DNS_R_FORMERR,
 				      "meta-RR in update");
 		} else {
-			isc_log_write(ns_g_lctx, NS_LOGCATEGORY_UPDATE,
-				      NS_LOGMODULE_UPDATE, ISC_LOG_WARNING,
-				      "update RR has incorrect class %d",
-				      update_class);
+			update_log(client, zone, ISC_LOG_WARNING,
+				   "update RR has incorrect class %d",
+				   update_class);
 			FAIL(DNS_R_FORMERR);
 		}
 		/*
@@ -2192,7 +2225,8 @@ update_action(isc_task_t *task, isc_event_t *event) {
 	if (result != ISC_R_NOMORE)
 		FAIL(result);
 
-	isc_log_write(UPDATE_DEBUG_LOGARGS, "update section prescan OK");
+	update_log(client, zone, LOGLEVEL_DEBUG,
+		   "update section prescan OK");
 
 	/*
 	 * Process the Update Section.
@@ -2217,10 +2251,11 @@ update_action(isc_task_t *task, isc_event_t *event) {
 								      name,
 								      &flag));
 				if (flag) {
-					isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-						      "attempt to add CNAME "
-						      "alongside non-CNAME "
-						      "ignored");
+					update_log(client, zone,
+						   LOGLEVEL_PROTOCOL,
+						   "attempt to add CNAME "
+						   "alongside non-CNAME "
+						   "ignored");
 					continue;
 				}
 			} else {
@@ -2230,9 +2265,10 @@ update_action(isc_task_t *task, isc_event_t *event) {
 				if (flag &&
 				    ! dns_rdatatype_isdnssec(rdata.type))
 				{
-					isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-					      "attempt to add non-CNAME "
-					      "alongside CNAME ignored");
+					update_log(client, zone,
+						   LOGLEVEL_PROTOCOL,
+						   "attempt to add non-CNAME "
+						   "alongside CNAME ignored");
 					continue;
 				}
 			}
@@ -2242,24 +2278,27 @@ update_action(isc_task_t *task, isc_event_t *event) {
 						   dns_rdatatype_soa, 0,
 						   &flag));
 				if (! flag) {
-					isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-						      "attempt to create 2nd "
-						      "SOA ignored");
+					update_log(client, zone,
+						   LOGLEVEL_PROTOCOL,
+						   "attempt to create 2nd "
+						   "SOA ignored");
 					continue;
 				}
 				CHECK(check_soa_increment(db, ver, &rdata,
 							  &ok));
 				if (! ok) {
-					isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-						      "SOA update failed to "
-						      "increment serial, "
-						      "ignoring it");
+					update_log(client, zone,
+						   LOGLEVEL_PROTOCOL,
+						   "SOA update failed to "
+						   "increment serial, "
+						   "ignoring it");
 					continue;
 				}
 				soa_serial_changed = ISC_TRUE;
 			}
 			
-			isc_log_write(UPDATE_PROTOCOL_LOGARGS, "adding an RR");
+			update_log(client, zone,
+				   LOGLEVEL_PROTOCOL, "adding an RR");
 
 			/* Prepare the affected RRset for the addition. */
 			{
@@ -2289,8 +2328,9 @@ update_action(isc_task_t *task, isc_event_t *event) {
 			}
 		} else if (update_class == dns_rdataclass_any) {
 			if (rdata.type == dns_rdatatype_any) {
-				isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-					      "delete all rrsets from a name");
+				update_log(client, zone,
+					   LOGLEVEL_PROTOCOL,
+					   "delete all rrsets from a name");
 				if (dns_name_equal(name, zonename)) {
 					CHECK(delete_if(type_not_soa_nor_ns_p,
 							db, ver, name,
@@ -2304,13 +2344,15 @@ update_action(isc_task_t *task, isc_event_t *event) {
 			} else if (dns_name_equal(name, zonename) &&
 				   (rdata.type == dns_rdatatype_soa ||
 				    rdata.type == dns_rdatatype_ns)) {
-				isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-					      "attempt to delete all SOA "
-					      "or NS records ignored");
+				update_log(client, zone,
+					   LOGLEVEL_PROTOCOL,
+					   "attempt to delete all SOA "
+					   "or NS records ignored");
 				continue;
 			} else {
-				isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-					      "deleting an rrset");
+				update_log(client, zone,
+					   LOGLEVEL_PROTOCOL,
+					   "deleting an rrset");
 				CHECK(delete_if(true_p, db, ver, name,
 						rdata.type, covers, &rdata,
 						&diff));
@@ -2322,9 +2364,10 @@ update_action(isc_task_t *task, isc_event_t *event) {
 			 */
 			if (dns_name_equal(name, zonename)) {
 				if (rdata.type == dns_rdatatype_soa) {
-					isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-						      "attempt to delete SOA "
-						      "ignored");
+					update_log(client, zone,
+						   LOGLEVEL_PROTOCOL,
+						   "attempt to delete SOA "
+						   "ignored");
 					continue;
 				}
 				if (rdata.type == dns_rdatatype_ns) {
@@ -2333,16 +2376,18 @@ update_action(isc_task_t *task, isc_event_t *event) {
 						       dns_rdatatype_ns,
 						       0, &count));
 					if (count == 1) {
-						isc_log_write(
-						UPDATE_PROTOCOL_LOGARGS,
-						      "attempt to delete last "
-						      "NS ignored");
+						update_log(client, zone,
+							   LOGLEVEL_PROTOCOL,
+							   "attempt to "
+							   "delete last "
+							   "NS ignored");
 						continue;
 					}
 				}
 			}
-			isc_log_write(UPDATE_PROTOCOL_LOGARGS,
-				      "deleting an RR");
+			update_log(client, zone,
+				   LOGLEVEL_PROTOCOL,
+				   "deleting an RR");
 			CHECK(delete_if(rr_equal_p, db, ver, name,
 					rdata.type, covers, &rdata, &diff));
 		}
@@ -2368,23 +2413,21 @@ update_action(isc_task_t *task, isc_event_t *event) {
 		}
 
 		if (dns_db_issecure(db)) {
-
 			result = update_signatures(mctx, db, oldver, ver,
 			   &diff, dns_zone_getsigvalidityinterval(zone));
 			if (result != ISC_R_SUCCESS) {
-				isc_log_write(ns_g_lctx, NS_LOGCATEGORY_UPDATE,
-					      NS_LOGMODULE_UPDATE,
-					      ISC_LOG_ERROR,
-					      "SIG/NXT update failed: %s",
-					      isc_result_totext(result));
+				update_log(client, zone,
+					   ISC_LOG_ERROR,
+					   "SIG/NXT update failed: %s",
+					   isc_result_totext(result));
 				goto failure;
 			}
 		}
 
 		journalfile = dns_zone_getjournal(zone);
 		if (journalfile != NULL) {
-			isc_log_write(UPDATE_DEBUG_LOGARGS,
-				      "writing journal %s", journalfile);
+			update_log(client, zone, LOGLEVEL_DEBUG,
+				   "writing journal %s", journalfile);
 
 			journal = NULL;
 			result = dns_journal_open(mctx, journalfile,
@@ -2407,7 +2450,8 @@ update_action(isc_task_t *task, isc_event_t *event) {
 	 *         to handle databases that need two-phase commit, but this
 	 *	   isn't a priority.
 	 */
-	isc_log_write(UPDATE_DEBUG_LOGARGS, "committing update transaction");
+	update_log(client, zone, LOGLEVEL_DEBUG,
+		   "committing update transaction");
 	dns_db_closeversion(db, &ver, ISC_TRUE);
 
 	/*
@@ -2419,7 +2463,6 @@ update_action(isc_task_t *task, isc_event_t *event) {
 	 * Notify slaves of the change we just made.
 	 */
 	dns_zone_notify(zone);
-
 	result = ISC_R_SUCCESS;
 	goto common;
 
@@ -2428,8 +2471,8 @@ update_action(isc_task_t *task, isc_event_t *event) {
 	 * The reason for failure should have been logged at this point.
 	 */
 	if (ver != NULL) {
-		isc_log_write(UPDATE_DEBUG_LOGARGS,
-			      "rolling back");
+		update_log(client, zone, LOGLEVEL_DEBUG, 
+			   "rolling back");
 		dns_db_closeversion(db, &ver, ISC_FALSE);
 	}
 
