@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2001  Internet Software Consortium.
+ * Copyright (C) 1999-2002  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -15,7 +15,7 @@
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: view.c,v 1.89.2.1.10.2 2003/09/18 00:58:03 marka Exp $ */
+/* $Id: view.c,v 1.103.2.2.6.2 2003/09/18 01:00:04 marka Exp $ */
 
 #include <config.h>
 
@@ -81,14 +81,6 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 		result = ISC_R_UNEXPECTED;
 		goto cleanup_name;
 	}
-	result = isc_rwlock_init(&view->conflock, 1, 1);
-	if (result != ISC_R_SUCCESS) {
-		UNEXPECTED_ERROR(__FILE__, __LINE__,
-				 "isc_rwlock_init() failed: %s",
-				 isc_result_totext(result));
-		result = ISC_R_UNEXPECTED;
-		goto cleanup_mutex;
-	}
 	view->zonetable = NULL;
 	result = dns_zt_create(mctx, rdclass, &view->zonetable);
 	if (result != ISC_R_SUCCESS) {
@@ -96,7 +88,7 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 				 "dns_zt_create() failed: %s",
 				 isc_result_totext(result));
 		result = ISC_R_UNEXPECTED;
-		goto cleanup_rwlock;
+		goto cleanup_mutex;
 	}
 	view->secroots = NULL;
 	result = dns_keytable_create(mctx, &view->secroots);
@@ -131,19 +123,20 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 	view->hints = NULL;
 	view->resolver = NULL;
 	view->adb = NULL;
-	view->loadmgr = NULL;
 	view->requestmgr = NULL;
 	view->mctx = mctx;
 	view->rdclass = rdclass;
 	view->frozen = ISC_FALSE;
 	view->task = NULL;
-	view->references = 1;
+	isc_refcount_init(&view->references, 1);
 	view->weakrefs = 0;
 	view->attributes = (DNS_VIEWATTR_RESSHUTDOWN|DNS_VIEWATTR_ADBSHUTDOWN|
 			    DNS_VIEWATTR_REQSHUTDOWN);
 	view->statickeys = NULL;
 	view->dynamickeys = NULL;
 	view->matchclients = NULL;
+	view->matchdestinations = NULL;
+	view->matchrecursiveonly = ISC_FALSE;
 	result = dns_tsigkeyring_create(view->mctx, &view->dynamickeys);
 	if (result != ISC_R_SUCCESS)
 		goto cleanup_fwdtable;
@@ -157,20 +150,27 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 	view->auth_nxdomain = ISC_FALSE; /* Was true in BIND 8 */
 	view->additionalfromcache = ISC_TRUE;
 	view->additionalfromauth = ISC_TRUE;
+	view->minimalresponses = ISC_FALSE;
 	view->transfer_format = dns_one_answer;
 	view->queryacl = NULL;
 	view->recursionacl = NULL;
+	view->v6synthesisacl = NULL;
 	view->sortlist = NULL;
 	view->requestixfr = ISC_TRUE;
 	view->provideixfr = ISC_TRUE;
 	view->maxcachettl = 7 * 24 * 3600;
 	view->maxncachettl = 3 * 3600;
 	view->dstport = 53;
-	view->cachefile = NULL;
+	view->flush = ISC_FALSE;
 
 	result = dns_peerlist_new(view->mctx, &view->peers);
 	if (result != ISC_R_SUCCESS)
 		goto cleanup_dynkeys;
+
+	result = dns_aclenv_init(view->mctx, &view->aclenv);
+	if (result != ISC_R_SUCCESS)
+		goto cleanup_peerlist;
+
 	ISC_LINK_INIT(view, link);
 	ISC_EVENT_INIT(&view->resevent, sizeof view->resevent, 0, NULL,
 		       DNS_EVENT_VIEWRESSHUTDOWN, resolver_shutdown,
@@ -187,6 +187,9 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 
 	return (ISC_R_SUCCESS);
 
+ cleanup_peerlist:
+	dns_peerlist_detach(&view->peers);
+
  cleanup_dynkeys:
 	dns_tsigkeyring_destroy(&view->dynamickeys);
 
@@ -202,9 +205,6 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
  cleanup_zt:
 	dns_zt_detach(&view->zonetable);
 
- cleanup_rwlock:
-	isc_rwlock_destroy(&view->conflock);
-
  cleanup_mutex:
 	DESTROYLOCK(&view->lock);
 
@@ -220,7 +220,7 @@ dns_view_create(isc_mem_t *mctx, dns_rdataclass_t rdclass,
 static inline void
 destroy(dns_view_t *view) {
 	REQUIRE(!ISC_LINK_LINKED(view, link));
-	REQUIRE(view->references == 0);
+	REQUIRE(isc_refcount_current(&view->references) == 0);
 	REQUIRE(view->weakrefs == 0);
 	REQUIRE(RESSHUTDOWN(view));
 	REQUIRE(ADBSHUTDOWN(view));
@@ -236,8 +236,6 @@ destroy(dns_view_t *view) {
 		dns_adb_detach(&view->adb);
 	if (view->resolver != NULL)
 		dns_resolver_detach(&view->resolver);
-	if (view->loadmgr != NULL)
-		dns_loadmgr_detach(&view->loadmgr);
 	if (view->requestmgr != NULL)
 		dns_requestmgr_detach(&view->requestmgr);
 	if (view->task != NULL)
@@ -250,14 +248,16 @@ destroy(dns_view_t *view) {
 		dns_cache_detach(&view->cache);
 	if (view->matchclients != NULL)
 		dns_acl_detach(&view->matchclients);
+	if (view->matchdestinations != NULL)
+		dns_acl_detach(&view->matchdestinations);
 	if (view->queryacl != NULL)
 		dns_acl_detach(&view->queryacl);
 	if (view->recursionacl != NULL)
 		dns_acl_detach(&view->recursionacl);
+	if (view->v6synthesisacl != NULL)
+		dns_acl_detach(&view->v6synthesisacl);
 	if (view->sortlist != NULL)
 		dns_acl_detach(&view->sortlist);
-	if (view->cachefile != NULL)
-		isc_mem_free(view->mctx, view->cachefile);
 	if (view->delonly != NULL) {
 		dns_name_t *name;
 		int i;
@@ -278,8 +278,9 @@ destroy(dns_view_t *view) {
 	dns_keytable_detach(&view->trustedkeys);
 	dns_keytable_detach(&view->secroots);
 	dns_fwdtable_destroy(&view->fwdtable);
-	isc_rwlock_destroy(&view->conflock);
+	dns_aclenv_destroy(&view->aclenv);
 	DESTROYLOCK(&view->lock);
+	isc_refcount_destroy(&view->references);
 	isc_mem_free(view->mctx, view->name);
 	isc_mem_put(view->mctx, view, sizeof *view);
 }
@@ -291,7 +292,8 @@ destroy(dns_view_t *view) {
 static isc_boolean_t
 all_done(dns_view_t *view) {
 
-	if (view->references == 0 && view->weakrefs == 0 &&
+	if (isc_refcount_current(&view->references) == 0 &&
+	    view->weakrefs == 0 &&
 	    RESSHUTDOWN(view) && ADBSHUTDOWN(view) && REQSHUTDOWN(view))
 		return (ISC_TRUE);
 
@@ -304,13 +306,7 @@ dns_view_attach(dns_view_t *source, dns_view_t **targetp) {
 	REQUIRE(DNS_VIEW_VALID(source));
 	REQUIRE(targetp != NULL && *targetp == NULL);
 
-	LOCK(&source->lock);
-
-	INSIST(source->references > 0);
-	source->references++;
-	INSIST(source->references != 0);
-
-	UNLOCK(&source->lock);
+	isc_refcount_increment(&source->references, NULL);
 
 	*targetp = source;
 }
@@ -318,30 +314,31 @@ dns_view_attach(dns_view_t *source, dns_view_t **targetp) {
 static void
 view_flushanddetach(dns_view_t **viewp, isc_boolean_t flush) {
 	dns_view_t *view;
+	unsigned int refs;
 	isc_boolean_t done = ISC_FALSE;
 
 	REQUIRE(viewp != NULL);
 	view = *viewp;
 	REQUIRE(DNS_VIEW_VALID(view));
 
-	LOCK(&view->lock);
-
-	INSIST(view->references > 0);
-	view->references--;
-	if (view->references == 0) {
+	if (flush)
+		view->flush = ISC_TRUE;
+	isc_refcount_decrement(&view->references, &refs);
+	if (refs == 0) {
+		LOCK(&view->lock);
 		if (!RESSHUTDOWN(view))
 			dns_resolver_shutdown(view->resolver);
 		if (!ADBSHUTDOWN(view))
 			dns_adb_shutdown(view->adb);
 		if (!REQSHUTDOWN(view))
 			dns_requestmgr_shutdown(view->requestmgr);
-		if (flush)
+		if (view->flush)
 			dns_zt_flushanddetach(&view->zonetable);
 		else
 			dns_zt_detach(&view->zonetable);
 		done = all_done(view);
+		UNLOCK(&view->lock);
 	}
-	UNLOCK(&view->lock);
 
 	*viewp = NULL;
 
@@ -581,15 +578,6 @@ dns_view_setdstport(dns_view_t *view, in_port_t dstport) {
 	view->dstport = dstport;
 }
 
-void
-dns_view_setloadmgr(dns_view_t *view, dns_loadmgr_t *loadmgr) {
-	REQUIRE(DNS_VIEW_VALID(view));
-
-	if (view->loadmgr != NULL)
-		dns_loadmgr_detach(&view->loadmgr);
-	dns_loadmgr_attach(loadmgr, &view->loadmgr);
-}
-
 isc_result_t
 dns_view_addzone(dns_view_t *view, dns_zone_t *zone) {
 	isc_result_t result;
@@ -784,6 +772,12 @@ dns_view_find(dns_view_t *view, dns_name_t *name, dns_rdatatype_t type,
 			result = DNS_R_HINTNXRRSET;
 		} else if (result == DNS_R_NXDOMAIN)
 			result = ISC_R_NOTFOUND;
+
+		/*
+		 * Cleanup if non-standard hints are used.
+		 */
+		if (db == NULL && node != NULL)
+			dns_db_detachnode(view->hints, &node);
 	}
 
  cleanup:
@@ -821,7 +815,8 @@ dns_view_find(dns_view_t *view, dns_name_t *name, dns_rdatatype_t type,
 			*dbp = db;
 		else
 			dns_db_detach(&db);
-	}
+	} else
+		INSIST(node == NULL);
 
 	if (zone != NULL)
 		dns_zone_detach(&zone);
@@ -951,8 +946,7 @@ dns_view_findzonecut(dns_view_t *view, dns_name_t *name, dns_name_t *fname,
 			 * We found an answer, but the cache may be better.
 			 */
 			zfname = dns_fixedname_name(&zfixedname);
-			result = dns_name_concatenate(fname, NULL, zfname,
-						      NULL);
+			result = dns_name_copy(fname, zfname, NULL);
 			if (result != ISC_R_SUCCESS)
 				goto cleanup;
 			dns_rdataset_clone(rdataset, &zrdataset);
@@ -1008,7 +1002,7 @@ dns_view_findzonecut(dns_view_t *view, dns_name_t *name, dns_name_t *fname,
 			    dns_rdataset_isassociated(sigrdataset))
 				dns_rdataset_disassociate(sigrdataset);
 		}
-		result = dns_name_concatenate(zfname, NULL, fname, NULL);
+		result = dns_name_copy(zfname, fname, NULL);
 		if (result != ISC_R_SUCCESS)
 			goto cleanup;
 		dns_rdataset_clone(&zrdataset, rdataset);
@@ -1027,6 +1021,8 @@ dns_view_findzonecut(dns_view_t *view, dns_name_t *name, dns_name_t *fname,
 			 * We can't even find the hints for the root
 			 * nameservers!
 			 */
+			if (dns_rdataset_isassociated(rdataset))
+				dns_rdataset_disassociate(rdataset);
 			result = ISC_R_NOTFOUND;
 		}
 	}
@@ -1075,6 +1071,14 @@ dns_view_load(dns_view_t *view, isc_boolean_t stop) {
 	return (dns_zt_load(view->zonetable, stop));
 }
 
+isc_result_t
+dns_view_loadnew(dns_view_t *view, isc_boolean_t stop) {
+
+	REQUIRE(DNS_VIEW_VALID(view));
+
+	return (dns_zt_loadnew(view->zonetable, stop));
+}
+
 isc_result_t
 dns_view_gettsig(dns_view_t *view, dns_name_t *keyname, dns_tsigkey_t **keyp)
 {
@@ -1117,17 +1121,6 @@ dns_view_checksig(dns_view_t *view, isc_buffer_t *source, dns_message_t *msg) {
 				view->dynamickeys));
 }
 
-isc_result_t
-dns_view_dumpcache(dns_view_t *view) {
-	REQUIRE(DNS_VIEW_VALID(view));
-
-	if (view->cachefile == NULL)
-		return (ISC_R_IGNORE);
-	return (dns_master_dump(view->mctx, view->cachedb, NULL,
-				&dns_master_style_default, view->cachefile));
-
-}
-
 isc_result_t
 dns_view_dumpdbtostream(dns_view_t *view, FILE *fp) {
 	isc_result_t result;
@@ -1136,7 +1129,7 @@ dns_view_dumpdbtostream(dns_view_t *view, FILE *fp) {
 
 	(void)fprintf(fp, ";\n; Cache dump of view '%s'\n;\n", view->name);
 	result = dns_master_dumptostream(view->mctx, view->cachedb, NULL,
-					  &dns_master_style_explicitttl, fp);
+					  &dns_master_style_cache, fp);
 	if (result != ISC_R_SUCCESS)
 		return (result);
 #ifdef notyet /* clean up adb dump format first */
@@ -1145,6 +1138,24 @@ dns_view_dumpdbtostream(dns_view_t *view, FILE *fp) {
 	return (ISC_R_SUCCESS);
 }
 
+isc_result_t
+dns_view_flushcache(dns_view_t *view) {
+	isc_result_t result;
+
+	REQUIRE(DNS_VIEW_VALID(view));
+
+	if (view->cachedb == NULL)
+		return (ISC_R_SUCCESS);
+	result = dns_cache_flush(view->cache);
+	if (result != ISC_R_SUCCESS)
+		return (result);
+	dns_db_detach(&view->cachedb);
+	dns_cache_attachdb(view->cache, &view->cachedb);
+
+	dns_adb_flush(view->adb);
+	return (ISC_R_SUCCESS);
+}
+
 isc_result_t
 dns_view_adddelegationonly(dns_view_t *view, dns_name_t *name) {
 	isc_result_t result;
